/* pydiacanvasitem.c
 * Copyright (C) 2002  Arjan Molenaar
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <diacanvas/dia-canvas.h>
#include <diacanvas/diatypebuiltins.h>
#include <pygobject.h>

//#define ENABLE_DEBUG 1
//#define ENABLE_DEBUG_GROUP 1

#define METHOD_PREFIX "on_"
#define DATA_IS_PYOBJECT ((int*) 1)
#define DATA_IS_NOT_PYOBJECT ((int*) 2)

extern PyTypeObject PyDiaHandle_Type;
extern PyTypeObject PyDiaCanvasItem_Type;

static gboolean
iter_from_object (DiaCanvasIter *iter, PyObject *obj)
{
	if (obj) {
		if (pyg_boxed_check(obj, DIA_TYPE_CANVAS_ITER)) {
			DiaCanvasIter *py_iter = pyg_boxed_get (obj, DiaCanvasIter);
			*iter = *py_iter;
			iter->data[2] = DATA_IS_NOT_PYOBJECT;
			Py_DECREF (obj);
			return TRUE;
		} else if (obj != Py_None) {
			//g_message (__FUNCTION__": Data is a PyObject!");
			iter->data[0] = obj;
			iter->data[2] = DATA_IS_PYOBJECT;
			Py_DECREF (obj);
			return TRUE;
		} else {
			iter->data[0] = NULL;
			Py_DECREF (obj);
			return FALSE;
		}
	}
	iter->data[0] = NULL;
	PyErr_Print ();
	PyErr_Clear ();
	return FALSE;
}

static gboolean
iter_as_object (DiaCanvasIter *iter, PyObject **obj)
{
	gboolean unref_data = FALSE;

	if (iter->data[2] == DATA_IS_PYOBJECT) {
		*obj = iter->data[0];
	} else if (iter->data[2] == DATA_IS_NOT_PYOBJECT) {
		*obj = pyg_boxed_new (DIA_TYPE_CANVAS_ITER, iter, FALSE, FALSE);
		unref_data = TRUE;
	} else {
		g_assert_not_reached();
	}
	return unref_data;
}

/*
 * DiaCanvasItem
 */
void	pydia_canvas_item_class_init	(DiaCanvasItemClass *klass);

static void	pydia_canvas_item_update		(DiaCanvasItem *item,
							 gdouble affine[6]);
static gboolean pydia_canvas_item_get_shape_iter	(DiaCanvasItem *item,
							 DiaCanvasIter *iter);
static gboolean pydia_canvas_item_shape_next		(DiaCanvasItem *item,
							 DiaCanvasIter *iter);
static DiaShape* pydia_canvas_item_shape_value		(DiaCanvasItem *item,
							 DiaCanvasIter *iter);
static gdouble	pydia_canvas_item_point			(DiaCanvasItem *item,
							 gdouble x, gdouble y);
static void	pydia_canvas_item_handle_motion		(DiaCanvasItem *item,
							 DiaHandle *handle,
							 gdouble *wx, gdouble *wy,
							 DiaEventMask mask);
static gdouble	pydia_canvas_item_glue			(DiaCanvasItem *item,
							 DiaHandle *handle,
							 gdouble *wx,
							 gdouble *wy);
static gboolean pydia_canvas_item_event			(DiaCanvasItem *item,
							 DiaEvent *event);
static void	pydia_canvas_item_move			(DiaCanvasItem *item,
							 gdouble dx, gdouble dy,
							 gboolean interactive);
static gboolean pydia_canvas_item_connect		(DiaCanvasItem *item,
							 DiaHandle *handle);
static gboolean pydia_canvas_item_disconnect		(DiaCanvasItem *item,
							 DiaHandle *handle);

static void pydia_canvas_groupable_class_init (DiaCanvasGroupable *groupable);

/*
 * DiaCanvasItem
 */
void
pydia_canvas_item_class_init (DiaCanvasItemClass *klass)
{
	/* Callbacks */
	klass->update = pydia_canvas_item_update;
	klass->point = pydia_canvas_item_point;
	klass->handle_motion = pydia_canvas_item_handle_motion;
	klass->glue = pydia_canvas_item_glue;
	klass->get_shape_iter = pydia_canvas_item_get_shape_iter;
	klass->shape_next = pydia_canvas_item_shape_next;
	klass->shape_value = pydia_canvas_item_shape_value;

	/* Signals. */
	klass->event = pydia_canvas_item_event;
	klass->move = pydia_canvas_item_move;
	klass->connect = pydia_canvas_item_connect;
	klass->disconnect = pydia_canvas_item_disconnect;
}

static void
pydia_canvas_item_update (DiaCanvasItem *item, gdouble affine[6])
{
	PyObject *self;
	PyObject *py_affine, *py_ret;
	GValue v = { 0, };

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif

	pyg_block_threads();

	g_value_init (&v, DIA_TYPE_AFFINE);
	g_value_set_static_boxed (&v, affine);
	//py_affine = pyg_boxed_new (DIA_TYPE_AFFINE, affine, FALSE, FALSE);
	py_affine = pyg_value_as_pyobject (&v, TRUE);
	g_value_unset (&v);

	g_assert (py_affine != NULL);

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self, METHOD_PREFIX "update",
				      "(O)", py_affine);
	Py_DECREF (self);
	Py_DECREF (py_affine);
	if (!py_ret) {
		PyErr_Print();
		PyErr_Clear();
	} else
		Py_DECREF (py_ret);
	pyg_unblock_threads();
}

static gboolean
pydia_canvas_item_get_shape_iter (DiaCanvasItem *item, DiaCanvasIter *iter)
{
	PyObject *self;
	PyObject *py_ret;

#ifdef ENABLE_DEBUG2
	g_message (__FUNCTION__);
#endif

	pyg_block_threads();

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "get_shape_iter",
				      "");

	Py_DECREF (self);
	
	pyg_unblock_threads();

	/* Takes care of decrementing py_ret. */
	return iter_from_object (iter, py_ret);
}

static gboolean
pydia_canvas_item_shape_next (DiaCanvasItem *item, DiaCanvasIter *iter)
{
	PyObject *self;
	PyObject *py_ret;
	PyObject *data = NULL;
	gboolean unref_data = FALSE;

#ifdef ENABLE_DEBUG2
	g_message (__FUNCTION__);
#endif

	pyg_block_threads();

	unref_data = iter_as_object (iter, &data);

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "shape_next",
				      "(O)", data);
	Py_DECREF (self);
	if (unref_data)
		Py_DECREF ((PyObject*)data);

	pyg_unblock_threads();

	/* Takes care of decrementing py_ret. */
	return iter_from_object (iter, py_ret);
}

static DiaShape*
pydia_canvas_item_shape_value (DiaCanvasItem *item, DiaCanvasIter *iter)
{
	PyObject *self;
	DiaShape *shape = NULL;
	PyObject *py_ret;
	PyObject *data = NULL;
	gboolean unref_data = FALSE;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__": data=%p", iter->data[0]);
#endif

	pyg_block_threads();

	unref_data = iter_as_object (iter, &data);

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "shape_value",
				      "(O)", data);
	Py_DECREF (self);
	if (unref_data)
		Py_DECREF ((PyObject*) data);

	if (py_ret) {
		if (!(pyg_boxed_check (py_ret, DIA_TYPE_SHAPE_PATH)
		      || pyg_boxed_check (py_ret, DIA_TYPE_SHAPE_BEZIER)
		      || pyg_boxed_check (py_ret, DIA_TYPE_SHAPE_ELLIPSE)
		      || pyg_boxed_check (py_ret, DIA_TYPE_SHAPE_TEXT)
		      || pyg_boxed_check (py_ret, DIA_TYPE_SHAPE_IMAGE))) {
			PyErr_SetString (PyExc_TypeError,
				"return value should be a DiaShape.");
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return NULL;
		}
		//g_message ("getting shape");
		shape = pyg_boxed_get (py_ret, DiaShape);
		//g_message ("done");
		Py_DECREF (py_ret);
	} else {
		PyErr_Print();
		PyErr_Clear();
		pyg_unblock_threads();
		return NULL;
	}
	pyg_unblock_threads();
	return shape;
}

static gdouble
pydia_canvas_item_point (DiaCanvasItem *item, gdouble x, gdouble y)
{
	PyObject *self;
	PyObject *py_ret, *py_float;
	gdouble retval = G_MAXDOUBLE;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif
		
	pyg_block_threads();

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "point",
				      "(dd)", x, y);
	Py_DECREF (self);
	if (!py_ret) {
		PyErr_Clear ();
		PyErr_SetString (PyExc_TypeError,
				 "method " METHOD_PREFIX "point not found.");
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}
	py_float = PyNumber_Float(py_ret);
	if (py_float)
		retval = PyFloat_AsDouble (py_float);
	else {
		PyErr_Clear ();
		PyErr_SetString (PyExc_TypeError,
				 "return value should be a float.");
		Py_XDECREF (py_ret);
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}
	Py_DECREF (py_float);
	Py_DECREF (py_ret);

	pyg_unblock_threads();

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__": returning %f", retval);
#endif
	return retval;
}

static void
pydia_canvas_item_move (DiaCanvasItem *item, gdouble dx, gdouble dy, gboolean interactive)
{
	PyObject *self;
	PyObject *py_ret;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif
		
	pyg_block_threads();

	self = pygobject_new ((GObject*) item);

	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "move",
				      "(ddi)", dx, dy, interactive);
	Py_DECREF (self);


	if (!py_ret) {
		PyErr_Print();
		PyErr_Clear();
	} else
		Py_DECREF (py_ret);
	pyg_unblock_threads();
}

static void
pydia_canvas_item_handle_motion (DiaCanvasItem *item, DiaHandle *handle,
				 gdouble *wx, gdouble *wy, DiaEventMask mask)
{
	PyObject *self;
	PyObject *py_handle;
	PyObject *py_ret, *py_float, *sitem;
	gdouble val[2];
	int i;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif

	pyg_block_threads();

	self = pygobject_new ((GObject*) item);
	py_handle = pygobject_new ((GObject*) handle);
	
	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "handle_motion",
				      "(Oddi)", py_handle, *wx, *wy, mask);

	Py_DECREF (self);
	Py_DECREF (py_handle);

	if (!py_ret) {
		PyErr_Print();
		PyErr_Clear();
		pyg_unblock_threads();
		return;
	}
	if (PySequence_Length(py_ret) != 2) {
		PyErr_SetString(PyExc_TypeError,
				"return value must be a 2 tuple of floats.");
		Py_DECREF (py_ret);
		pyg_unblock_threads();
		return;
	}

	for (i = 0; i < 2; i++) {
		sitem = PySequence_GetItem(py_ret, i);
		Py_DECREF(sitem);
		sitem = PyNumber_Float(sitem);
		if (sitem)
			val[i] = PyFloat_AsDouble(sitem);
		else {
			PyErr_Clear();
			PyErr_SetString(PyExc_TypeError,
					"sequence item not a float");
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return;
		}
		Py_DECREF(sitem);
	}
	*wx = val[0];
	*wy = val[1];
	Py_DECREF (py_ret);
	pyg_unblock_threads();
}

static gdouble
pydia_canvas_item_glue (DiaCanvasItem *item, DiaHandle *handle,
			gdouble *wx, gdouble *wy)
{
	PyObject *self;
	PyObject *py_handle, *py_ret, *py_wxy, *sitem;
	GValue v_wxy = { 0, };
	DiaPoint *wxy;
	gdouble retval = G_MAXDOUBLE;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif

	pyg_block_threads();

	self = pygobject_new ((GObject*) item);
	py_handle = pygobject_new ((GObject*)handle);
	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "glue",
				      "(Odd)", py_handle, *wx, *wy);
	Py_DECREF (self);
	Py_DECREF (py_handle);
	if (!py_ret || py_ret == Py_None) {
		Py_XDECREF (py_ret);
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}

	if (PySequence_Length(py_ret) != 2) {
		PyErr_SetString(PyExc_TypeError,
				"return value must be a tuple of a float and a point.");
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}

	/* Get distance */
	sitem = PySequence_GetItem(py_ret, 0);
	Py_DECREF(sitem);
	sitem = PyNumber_Float(sitem);
	if (sitem)
		retval = PyFloat_AsDouble(sitem);
	else {
		PyErr_Clear();
		PyErr_SetString(PyExc_TypeError,
				"first item of sequence item not a float");
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}
	Py_DECREF(sitem);

	/* Get glue point */
	g_value_init (&v_wxy, DIA_TYPE_POINT);

	sitem = PySequence_GetItem(py_ret, 1);

	if (pyg_value_from_pyobject(&v_wxy, sitem) != 0) {
		PyErr_SetString(PyExc_TypeError, "second item of sequence is not a tuple of 2 floats");
		Py_DECREF(sitem);
		pyg_unblock_threads();
		return G_MAXDOUBLE;
	}
	wxy = g_value_get_boxed (&v_wxy);

	*wx = wxy->x;
	*wy = wxy->y;
	g_value_unset (&v_wxy);
	Py_DECREF(sitem);
	Py_DECREF(py_ret);

	pyg_unblock_threads();

	return retval;
}

static gboolean
pydia_canvas_item_event (DiaCanvasItem *item, DiaEvent *event)
{
	PyObject *self;
	PyObject *py_event;
	PyObject *py_ret, *py_int;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif
		
	pyg_block_threads();

	self = pygobject_new ((GObject*) item);
	py_event = pyg_boxed_new (DIA_TYPE_EVENT, event, FALSE, FALSE);
	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "event",
				      "(O)", py_event);
	Py_DECREF (self);
	Py_DECREF (py_event);
	if (py_ret) {
		if (py_ret != Py_None) {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return TRUE;
		} else {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return FALSE;
		}
	} else {
		PyErr_Print ();
		PyErr_Clear ();
	}
	pyg_unblock_threads();
	return FALSE;
}

static gboolean
pydia_canvas_item_connect (DiaCanvasItem *item, DiaHandle *handle)
{
	PyObject *self;
	PyObject *py_handle;
	PyObject *py_ret;

	self = pygobject_new ((GObject*) item);

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif
		
	pyg_block_threads();

	py_handle = pygobject_new ((GObject*)handle);
	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "connect_handle",
				      "(O)", py_handle);
	Py_DECREF (self);
	Py_DECREF (py_handle);
	if (py_ret) {
		if (py_ret != Py_None) {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return TRUE;
		} else {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return FALSE;
		}
	} else {
		PyErr_Print ();
		PyErr_Clear ();
	}
	pyg_unblock_threads();
	return FALSE;
}

static gboolean
pydia_canvas_item_disconnect (DiaCanvasItem *item, DiaHandle *handle)
{
	PyObject *self;
	PyObject *py_handle;
	PyObject *py_ret;

#ifdef ENABLE_DEBUG
	g_message (__FUNCTION__);
#endif
		
	pyg_block_threads();

	self = pygobject_new ((GObject*) item);
	py_handle = pygobject_new ((GObject*)handle);
	py_ret = PyObject_CallMethod (self,
				      METHOD_PREFIX "disconnect_handle",
				      "(O)", py_handle);
	Py_DECREF (self);
	Py_DECREF (py_handle);
	if (py_ret) {
		if (py_ret != Py_None) {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return TRUE;
		} else {
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return FALSE;
		}
	} else {
		PyErr_Print ();
		PyErr_Clear ();
	}
	pyg_unblock_threads();
	return FALSE;
}

/*
 * Methods called by the python code to call their superclass.
 */

PyObject *
pydia_canvas_item_on_update(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "affine", NULL };
	PyObject *py_affine;
	GValue v = { 0, };
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O:DiaCanvasItem.on_update", kwlist, &py_affine))
		return NULL;
	g_value_init (&v, DIA_TYPE_AFFINE);

	if (pyg_value_from_pyobject (&v, py_affine) != 0)
		return NULL;
    
	klass = g_type_class_ref (type);
	if (klass->update)
		klass->update (DIA_CANVAS_ITEM(self->obj),
			       g_value_get_boxed (&v));
	g_type_class_unref (klass);

	g_value_unset (&v);

	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *
pydia_canvas_item_on_get_shape_iter(PyGObject *self, GType type)
{
	gboolean ret = FALSE;
	DiaCanvasIter iter = { NULL, NULL, NULL, 0 };
	DiaCanvasItemClass *klass;

	klass = g_type_class_ref (type);
	if (klass->get_shape_iter)
		ret = klass->get_shape_iter (DIA_CANVAS_ITEM(self->obj), &iter);
	g_type_class_unref (klass);

	if (ret) {
		PyObject *py_ret = pyg_boxed_new (DIA_TYPE_CANVAS_ITER,
						  &iter, TRUE, TRUE);
		return py_ret;
	}
	Py_INCREF (Py_None);
	return Py_None;
}

PyObject *
pydia_canvas_item_on_shape_next(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "iter", NULL };
	PyObject *py_iter;
	gboolean ret = FALSE;
	DiaCanvasIter *iter = NULL;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "O:DiaCanvasItem.on_shape_next",
					 kwlist, &py_iter))
		return NULL;
	if (pyg_boxed_check(py_iter, DIA_TYPE_CANVAS_ITER))
		iter = pyg_boxed_get(py_iter, DiaCanvasIter);
	else {
		Py_INCREF (Py_None);
		return Py_None;
	}
	
	//ret = dia_canvas_item_on_shape_next(DIA_CANVAS_ITEM(self->obj), iter);
	klass = g_type_class_ref (type);
	if (klass->shape_next)
		ret = klass->shape_next (DIA_CANVAS_ITEM(self->obj), iter);
	g_type_class_unref (klass);

	if (ret) {
		//PyObject *py_ret = pyg_boxed_new (DIA_TYPE_CANVAS_ITER, &iter, TRUE, TRUE);
		//return py_ret;
		Py_INCREF (py_iter);
		return py_iter;
	}

	Py_INCREF (Py_None);
	return Py_None;
}

PyObject *
pydia_canvas_item_on_shape_value(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "iter", NULL };
	PyObject *py_iter;
	DiaCanvasIter *iter = NULL;
	DiaShape *ret = NULL;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "O:DiaCanvasItem.on_shape_value",
					 kwlist, &py_iter))
		return NULL;
	if (pyg_boxed_check(py_iter, DIA_TYPE_CANVAS_ITER))
		iter = pyg_boxed_get(py_iter, DiaCanvasIter);
	else {
		Py_INCREF (Py_None);
		return Py_None;
	}

	klass = g_type_class_ref (type);
	if (klass->shape_value)
		ret = klass->shape_value (DIA_CANVAS_ITEM(self->obj), iter);
	g_type_class_unref (klass);

	if (ret)
		return pyg_boxed_new(dia_shape_get_type (ret), ret, FALSE, FALSE);
	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *
pydia_canvas_item_on_point(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "x", "y", NULL };
	double x, y, ret;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "dd:DiaCanvasItem.on_point",
					 kwlist, &x, &y))
		return NULL;

	klass = g_type_class_ref (type);
	if (klass->point)
		ret = klass->point (DIA_CANVAS_ITEM(self->obj), x, y);
	g_type_class_unref (klass);

	return PyFloat_FromDouble(ret);
}

PyObject *
pydia_canvas_item_on_move(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "x", "y", "interactive", NULL };
	double x, y;
	int interactive;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "ddi:DiaCanvasItem.on_move",
					 kwlist, &x, &y, &interactive))
		return NULL;

	klass = g_type_class_ref (type);
	if (klass->move)
		klass->move (DIA_CANVAS_ITEM(self->obj), x, y,
			     (gboolean) interactive);
	g_type_class_unref (klass);

	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *
pydia_canvas_item_on_handle_motion(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "handle", "x", "y", "mask", NULL };
	PyGObject *handle;
	double x, y;
	DiaEventMask mask;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "Oddi:DiaCanvasItem.on_handle_motion",
					 kwlist, &handle, &x, &y, &mask))
		return NULL;
	if (!pygobject_check(handle, &PyDiaHandle_Type)) {
		PyErr_SetString(PyExc_TypeError, "handle should be a DiaHandle");
		return NULL;
	}

	klass = g_type_class_ref (type);
	if (klass->handle_motion)
		klass->handle_motion (DIA_CANVAS_ITEM(self->obj), DIA_HANDLE (handle->obj), &x, &y, mask);
	g_type_class_unref (klass);

	/* We should somehow forward x and y. */
	return Py_BuildValue ("(dd)", x, y);
}

PyObject *
pydia_canvas_item_on_glue(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "handle", "x", "y", NULL };
	PyGObject *handle;
	double x = 0.0, y = 0.0, ret = G_MAXDOUBLE;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "Odd:DiaCanvasItem.on_glue",
					 kwlist, &handle, &x, &y))
		return NULL;
    
	klass = g_type_class_ref (type);
	if (klass->glue)
		ret = klass->glue (DIA_CANVAS_ITEM(self->obj), DIA_HANDLE (handle->obj), &x, &y);
	g_type_class_unref (klass);

	return Py_BuildValue ("(d(dd))", ret, x, y);
}

PyObject *
pydia_canvas_item_on_event(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "event", NULL };
	PyObject *py_event, *py_ret;
	int ret = FALSE;
	DiaEvent *event = NULL;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "O:DiaCanvasItem.on_event",
					 kwlist, &py_event))
		return NULL;
	if (pyg_boxed_check(py_event, DIA_TYPE_EVENT))
		event = pyg_boxed_get(py_event, DiaEvent);
	else {
		PyErr_SetString(PyExc_TypeError, "event should be a DiaEvent");
		return NULL;
	}
	//ret = dia_canvas_item_on_event(DIA_CANVAS_ITEM(self->obj), event);
	klass = g_type_class_ref (type);
	if (klass->event)
		ret = klass->event (DIA_CANVAS_ITEM(self->obj), event);
	g_type_class_unref (klass);

	py_ret = ret ? Py_True : Py_False;
	Py_INCREF(py_ret);
	return py_ret;
}

PyObject *
pydia_canvas_item_on_connect_handle(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "handle", NULL };
	PyGObject *handle;
	PyObject *py_ret;
	int ret = FALSE;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "O!:DiaCanvasItem.on_connect_handle",
					 kwlist, &PyDiaHandle_Type, &handle))
		return NULL;
	klass = g_type_class_ref (type);
	if (klass->connect)
		ret = klass->connect (DIA_CANVAS_ITEM(self->obj), DIA_HANDLE(handle->obj));
	g_type_class_unref (klass);

	py_ret = ret ? Py_True : Py_False;
	Py_INCREF(py_ret);
	return py_ret;
}

PyObject *
pydia_canvas_item_on_disconnect_handle(PyGObject *self, PyObject *args, PyObject *kwargs, GType type)
{
	static char *kwlist[] = { "handle", NULL };
	PyGObject *handle;
	PyObject *py_ret;
	int ret = FALSE;
	DiaCanvasItemClass *klass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs,
					 "O!:DiaCanvasItem.on_disconnect_handle",
					 kwlist, &PyDiaHandle_Type, &handle))
		return NULL;

	//ret = dia_canvas_item_on_diaconnect_handle(DIA_CANVAS_ITEM(self->obj), DIA_HANDLE(handle->obj));
	klass = g_type_class_ref (type);
	if (klass->disconnect)
		ret = klass->disconnect (DIA_CANVAS_ITEM(self->obj), DIA_HANDLE(handle->obj));
	g_type_class_unref (klass);

	py_ret = ret ? Py_True : Py_False;
	Py_INCREF(py_ret);
	return py_ret;
}

/*
 * DiaCanvasGroupable
 *
 * The only canvas item that is groupable by default is DiaCanvasGroup.
 *
 * For now only C->Python groupability will be coded.
 */

#undef METHOD_PREFIX
#define METHOD_PREFIX "on_groupable_"

static gboolean pydia_canvas_groupable_add		(DiaCanvasItem *group,
							 DiaCanvasItem *item);
static gboolean pydia_canvas_groupable_remove		(DiaCanvasItem *group,
							 DiaCanvasItem *item);
static gboolean pydia_canvas_groupable_get_iter		(DiaCanvasItem *group,
							 DiaCanvasIter *iter);
static gboolean pydia_canvas_groupable_next		(DiaCanvasItem *group,
							 DiaCanvasIter *iter);
static DiaCanvasItem* pydia_canvas_groupable_value	(DiaCanvasItem *group,
							 DiaCanvasIter *iter);
static gint	pydia_canvas_groupable_length		(DiaCanvasItem *group);
static gint	pydia_canvas_groupable_pos		(DiaCanvasItem *group,
							 DiaCanvasItem *item);

void
pydia_canvas_groupable_init (DiaCanvasGroupable *groupable)
{
	groupable->add = pydia_canvas_groupable_add;
	groupable->remove = pydia_canvas_groupable_remove;
	groupable->get_iter = pydia_canvas_groupable_get_iter;
	groupable->next = pydia_canvas_groupable_next;
	groupable->value = pydia_canvas_groupable_value;
	groupable->length = pydia_canvas_groupable_length;
	groupable->pos = pydia_canvas_groupable_pos;
}

static gboolean
pydia_canvas_groupable_add (DiaCanvasItem *group, DiaCanvasItem *item)
{
	PyObject *self;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif

	if (PyObject_HasAttrString (self, METHOD_PREFIX "add")) {
		PyObject *py_ret, *py_item;
		gboolean ret;

		pyg_block_threads();

		py_item = pygobject_new ((GObject*) item);
		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "add",
					      "O", py_item);
		Py_DECREF(py_item);
		Py_DECREF (self);
		if (py_ret) {
			ret = PyInt_AsLong(py_ret);
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return ret;
		} else {
			PyErr_Print ();
			PyErr_Clear ();
			pyg_unblock_threads();
			return FALSE;
		}
	}
	Py_DECREF (self);
	pyg_unblock_threads();
	return FALSE;
}

static gboolean
pydia_canvas_groupable_remove (DiaCanvasItem *group, DiaCanvasItem *item)
{
	PyObject *self;
	gboolean ret = FALSE;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "remove")) {
		PyObject *py_item, *py_ret;

		pyg_block_threads();

		py_item = pygobject_new ((GObject*) item);
		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "remove",
					      "O", py_item);
		Py_DECREF (py_item);
		if (py_ret) {
			ret = PyInt_AsLong(py_ret);
			Py_DECREF (py_ret);
		} else {
			Py_DECREF (self);
			PyErr_Print ();
			PyErr_Clear ();
			pyg_unblock_threads();
			return FALSE;
		}
		pyg_unblock_threads();
	}
	Py_DECREF (self);
	return ret;
}

static gboolean
pydia_canvas_groupable_get_iter (DiaCanvasItem *group, DiaCanvasIter *iter)
{
	PyObject *self;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "get_iter")) {
		PyObject *py_ret;

		pyg_block_threads();

		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "get_iter",
					      "");
		Py_DECREF (self);
		pyg_unblock_threads();

		/* Takes care of decrementing py_ret. */
		return iter_from_object (iter, py_ret);
	}
	Py_DECREF (self);
	return FALSE;
}

static gboolean
pydia_canvas_groupable_next (DiaCanvasItem *group, DiaCanvasIter *iter)
{
	PyObject *self;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "next")) {
		PyObject *py_ret;
		PyObject *data = NULL;
		gboolean unref_data = FALSE;

		pyg_block_threads();

		unref_data = iter_as_object (iter, &data);

		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "next",
					      "(O)", data);
		Py_DECREF (self);
		if (unref_data)
			Py_DECREF ((PyObject*)data);

		pyg_unblock_threads();

		return iter_from_object (iter, py_ret);
	}
	Py_DECREF (self);
	return FALSE;
}

static DiaCanvasItem*
pydia_canvas_groupable_value (DiaCanvasItem *group, DiaCanvasIter *iter)
{
	PyObject *self;
	DiaCanvasItem *item = NULL;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "value")) {
		PyObject *py_ret;
		PyObject *data = NULL;
		gboolean unref_data = FALSE;

		pyg_block_threads();

		unref_data = iter_as_object (iter, &data);

		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "value",
					      "(O)", data);
		Py_DECREF (self);
		if (unref_data)
			Py_DECREF ((PyObject*) data);
		if (py_ret) {
			if (!pygobject_check(py_ret, &PyDiaCanvasItem_Type)) {
				PyErr_SetString (PyExc_TypeError,
					"return value should be a DiaCanvasItem.");
				Py_DECREF (py_ret);
				pyg_unblock_threads();
				return NULL;
			}
			item =  DIA_CANVAS_ITEM (((PyGObject*)py_ret)->obj);
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return item;
		} else {
			PyErr_Print();
			PyErr_Clear();
			pyg_unblock_threads();
			return NULL;
		}
		pyg_unblock_threads();
	}
	Py_DECREF (self);
	return NULL;
}

static gint
pydia_canvas_groupable_length (DiaCanvasItem *group)
{
	PyObject *self;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "length")) {
		PyObject *py_ret;

		pyg_block_threads();

		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "length",
					      "");
		Py_DECREF (self);
		if (py_ret) {
			gint ret = PyInt_AsLong (py_ret);
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return ret;
		} else {
			PyErr_Print();
			PyErr_Clear();
			pyg_unblock_threads();
			return 0;
		}
		pyg_unblock_threads();
	}
	Py_DECREF (self);
	return 0;
}

static gint
pydia_canvas_groupable_pos (DiaCanvasItem *group, DiaCanvasItem *item)
{
	PyObject *self;

	self = pygobject_new ((GObject*) group);

#ifdef ENABLE_DEBUG_GROUP
	g_message (__FUNCTION__": self = %p, obj = %p", self, group);
#endif
	if (PyObject_HasAttrString (self, METHOD_PREFIX "pos")) {
		PyObject *py_ret, *py_item;

		pyg_block_threads();
		py_item = pygobject_new ((GObject*)item);
		py_ret = PyObject_CallMethod (self,
					      METHOD_PREFIX "pos",
					      "O", py_item);
		Py_DECREF (self);
		Py_DECREF (py_item);
		if (py_ret) {
			gint ret = PyInt_AsLong (py_ret);
			Py_DECREF (py_ret);
			pyg_unblock_threads();
			return ret;
		} else {
			PyErr_Print();
			PyErr_Clear();
			pyg_unblock_threads();
			return -1;
		}
		pyg_unblock_threads();
	}
	Py_DECREF (self);
	return -1;
}

