/* dia-handle-layer.c
 * Copyright (C) 2000  Arjan Molenaar
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "dia-handle-layer.h"
#include "dia-handle.h"
#include <libgnomecanvas/gnome-canvas-util.h>
#include <libart_lgpl/art_rgb_rgba_affine.h>
#include <math.h>
#include <string.h>
#include "dia-canvas-i18n.h"

//#define DEBUG

#ifdef DEBUG
# define D(msg) G_STMT_START { g_print (G_STRLOC": "); g_print msg; g_print ("\n"); } G_STMT_END
#else
# define D(msg)
#endif

#define DIA_HANDLE_EVENTS (GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_PRESS_MASK)

#define MAX_HANDLE_GLUE_DIST 10

enum {
	ARG_LAST
};

/* Stuff for handles and drawing of handles: */

/* We have the following criteria:
 * - Moveable on itself
 * - Connectable to other objects
 * - Is connected to another object
 */
enum {
	DIA_HANDLE_LAYER_MOVABLE = 1,
	DIA_HANDLE_LAYER_CONNECTABLE = 2,
	DIA_HANDLE_LAYER_IS_CONNECTED = 3,
	
	DIA_HANDLE_LAYER_FOCUS = 4
};


static guint32 _dia_handle_color[8] = {
	/* Just selected objects: */
	0x8080A0FF, /* 0 dark blue -> not moveable */
	0x80F080FF, /* 1 green -> moveable */
	0x80F080FF, /* 2 green w/ cross -> connectable */
	0xF08080FF, /* 3 red w/ cross -> connected */
	/* If object id focused: */
	0x0000D0FF, /* 4  */
	0x00D000FF, /* 5  */
	0x00D000FF, /* 6  */
	0xD00000FF  /* 7  */
};

static gchar *handle_image[8] = {
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

#define PIXEL(p, c) \
	G_STMT_START { \
		*(((char*) &p) + 0) = DIA_COLOR_RED (c); \
		*(((char*) &p) + 1) = DIA_COLOR_GREEN (c); \
		*(((char*) &p) + 2) = DIA_COLOR_BLUE (c); \
		*(((char*) &p) + 3) = DIA_COLOR_ALPHA (c); \
	} G_STMT_END

static void dia_handle_layer_init	(DiaHandleLayer *handle_layer);
static void dia_handle_layer_class_init (DiaHandleLayerClass  *klass);

/* GnomeCanvasItem signals */
static void dia_handle_layer_update     (GnomeCanvasItem *item,
					 double *affine,
					 ArtSVP *clip_path, int flags);
static void dia_handle_layer_realize    (GnomeCanvasItem *item);
static void dia_handle_layer_unrealize  (GnomeCanvasItem *item);
static void dia_handle_layer_map        (GnomeCanvasItem *item);
static void dia_handle_layer_unmap      (GnomeCanvasItem *item);
static void dia_handle_layer_draw       (GnomeCanvasItem *item,
					 GdkDrawable *drawable,
					 int x, int y,
					 int width, int height);
static void dia_handle_layer_render     (GnomeCanvasItem *item,
					 GnomeCanvasBuf *buf);
static double dia_handle_layer_point    (GnomeCanvasItem *item,
					 double x, double y,
					 int cx, int cy,
					 GnomeCanvasItem **actual_item);
static inline gboolean dia_handle_layer_button_press	(GnomeCanvasItem *item,
							 GdkEventButton *event);
static inline gboolean dia_handle_layer_button_release	(GnomeCanvasItem *item,
							 GdkEventButton *event);
static inline gboolean dia_handle_layer_motion_notify	(GnomeCanvasItem *item,
							 GdkEventMotion *event);
static gboolean dia_handle_layer_event	(GnomeCanvasItem *item,
					 GdkEvent *event);


static GnomeCanvasItemClass *parent_class = NULL;

/*
 * Some global initialization:
 */

/**
 * create_handle:
 * @handle_type: type of handle [0..7].
 *
 * Create a handle of size handle_size * handle_size.
 * The handles are created in RGBA format (4 bytes per pixel).
 *
 * Return value: A new handle
 **/
static gchar*
create_handle (int handle_type)
{
	guint32 *handle;
	gint i = 0; /* index */
	gint bufsize;
	gint handle_size;
	guint32 border;
	guint32 fill;
	guint32 cross;

	handle_size = dia_handle_size ();
	bufsize = handle_size * handle_size;

	PIXEL (border, 0x00000090);
	PIXEL (fill, _dia_handle_color[handle_type] & 0xFFFFFF90);
	PIXEL (cross, _dia_handle_color[handle_type] & 0x555555FF);

	handle = g_new (guint32, bufsize);
	
	g_assert (handle != NULL);

	/* Top and bottom row: */
	for ( ; i < handle_size; i++) {
		handle[i] = border;
		handle[bufsize - 1 - i] = border;
	}

	/* second row: */
	handle[i++] = border;
	for ( ; i < handle_size + handle_size - 1; i++)
		handle[i] = fill;
	handle[i++] = border;

	/* third row .. n-1 */
	for (; i < bufsize - handle_size; i += handle_size)
		memcpy (&handle[i], &handle[handle_size],
			handle_size * sizeof (guint32));

	/* If the handle can connect, draw a cross: */
	if (handle_type & DIA_HANDLE_LAYER_CONNECTABLE) {
		for (i = 2; i < handle_size - 2; i++) {
			/* \ */
			handle[i * handle_size + i] = cross;
			/* / */
			handle[i * handle_size + handle_size - i - 1] = cross;
		}
	}

	return (gchar*) handle;
}

/**
 * dia_handle_layer_create_images:
 *
 * Create images for the handles. This function should be called once from
 * dia_init().
 **/
static void
dia_handle_layer_create_images (void)
{
	int i;

	for (i = 0; i < 8; i++) {
		handle_image[i] = create_handle (i);
	}
}

/**
 * dia_handle_layer_free_images:
 *
 * This function is called via g_atexit().
 **/
static void
dia_handle_layer_free_images (void)
{
	int i;
	
	for (i = 0; i < 8; i++)
		g_free (handle_image[i]);
}


GtkType
dia_handle_layer_get_type (void)
{
	static GtkType handle_layer_type = 0;

	if (!handle_layer_type) {
		static const GtkTypeInfo handle_layer_info =
		{
			"DiaHandleLayer",
			sizeof (DiaHandleLayer),
			sizeof (DiaHandleLayerClass),
			(GtkClassInitFunc) dia_handle_layer_class_init,
			(GtkObjectInitFunc) dia_handle_layer_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
		
		handle_layer_type = gtk_type_unique (gnome_canvas_item_get_type (),
						    &handle_layer_info);
	}

	return handle_layer_type;
}

static void
dia_handle_layer_class_init (DiaHandleLayerClass *klass)
{
	GObjectClass *object_class;
	GnomeCanvasItemClass *item_class;
  
	/* Create an internal list of handles. */
	dia_handle_layer_create_images ();
	g_atexit (dia_handle_layer_free_images);

	object_class = (GObjectClass*) klass;
	item_class = (GnomeCanvasItemClass*) klass;
  
	parent_class = g_type_class_peek_parent (klass);

	item_class->update = dia_handle_layer_update;
	item_class->realize = dia_handle_layer_realize;
	item_class->unrealize = dia_handle_layer_unrealize;
	item_class->map = dia_handle_layer_map;
	item_class->unmap = dia_handle_layer_unmap;
	item_class->draw = dia_handle_layer_draw;
	item_class->render = dia_handle_layer_render;
	item_class->point = dia_handle_layer_point;
	item_class->event = dia_handle_layer_event;
}


static void
dia_handle_layer_init (DiaHandleLayer *hl)
{
	hl->glue_distance = 10;
	hl->grabbed_handle = NULL;
	hl->closest_handle = NULL;
	hl->closest_item = NULL;
	hl->cp_item = NULL;
	hl->event_mask = 0;
}


/* GnomeCanvasItem signals */
static void
dia_handle_layer_update (GnomeCanvasItem *item,
			 double *affine, ArtSVP *clip_path, int flags)
{
	D((" "));

        if (GNOME_CANVAS_ITEM_CLASS (parent_class)->update)
		GNOME_CANVAS_ITEM_CLASS (parent_class)->update (item, affine, clip_path, flags);

	DIA_HANDLE_LAYER (item)->glue_distance = MAX_HANDLE_GLUE_DIST / affine[0];
	gnome_canvas_request_redraw (item->canvas,
				     G_MININT, G_MININT,
				     G_MAXINT, G_MAXINT);

	item->x1 = item->y1 = G_MININT;
	item->x2 = item->y2 = G_MAXINT;
}

static void
dia_handle_layer_realize (GnomeCanvasItem *item)
{
	if (item->canvas->aa) {
		
	} else {
		/* create GC */
	}

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->realize)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->realize) (item);
}

static void
dia_handle_layer_unrealize (GnomeCanvasItem *item)
{

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->unrealize)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->unrealize) (item);
}

static void
dia_handle_layer_map (GnomeCanvasItem *item)
{

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->map)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->map) (item);
}

static void
dia_handle_layer_unmap (GnomeCanvasItem *item)
{

	if (GNOME_CANVAS_ITEM_CLASS (parent_class)->unmap)
		(* GNOME_CANVAS_ITEM_CLASS (parent_class)->unmap) (item);
}

static inline gint
find_handle_color (DiaHandle *handle, DiaCanvasViewItem *item)
{
	gint color;
	
	/* Figure out the handle color. */
	if (handle->connectable) {
		if (handle->connected_to)
			color = DIA_HANDLE_LAYER_IS_CONNECTED;
		else
			color = DIA_HANDLE_LAYER_CONNECTABLE;
	} else if (handle->movable) {
		color = DIA_HANDLE_LAYER_MOVABLE;
	} else
		color = 0;

	/* FixMe: Find out if the item has a child object that has the
	 * focus. The item should highlight if the item has DIA_COMPOSITE.
	 */

	/* Use highlight color if the object is focused: */
	if (DIA_CANVAS_VIEW_ITEM_FOCUS (item)
	    && GTK_WIDGET_HAS_FOCUS(GNOME_CANVAS_ITEM (item)->canvas)) {
		color |= DIA_HANDLE_LAYER_FOCUS;
	}
	return color;
}

typedef struct {
	GdkDrawable *drawable;
	int x, y, width, height;
} DrawHelper;

static gint
real_draw (DiaCanvasViewItem *item, gpointer data)
{
	DrawHelper *buf = data;
	DiaCanvasItem *diaitem = item->item;
	gdouble wx, wy, cx, cy;
	GList *l;
	//gdouble handle_r;
	gint handle_size;
	gint color;

	//D((G_GNUC_PRETTY_FUNCTION));

	/* Groups should draw their children... but invisible items and
	 * unselected leafs should not be drawn. */
	if ((!DIA_IS_CANVAS_GROUP (item->item)
	     && !DIA_CANVAS_VIEW_ITEM_SELECT (item))
	    || !(DIA_CANVAS_ITEM_VISIBLE (diaitem)))
		return FALSE;

	handle_size = dia_handle_size ();

	/* First draw all handles: */
	for (l = diaitem->handles; l != NULL; l = l->next) {
		DiaHandle *handle = l->data;

		color = find_handle_color (handle, item);

		dia_handle_get_pos_w (handle, &wx, &wy);
		gnome_canvas_w2c_d (GNOME_CANVAS_ITEM (item)->canvas,
				    wx, wy, &cx, &cy);

		gdk_draw_rgb_32_image (buf->drawable, 
				       item->gc,
				       cx - buf->x + 1, cy - buf->y + 1,
				       handle_size, handle_size,
				       GDK_RGB_DITHER_NORMAL,
				       handle_image[color],
				       handle_size * sizeof (guint32));
	}
	return TRUE;
}

static void
dia_handle_layer_draw (GnomeCanvasItem *hl, GdkDrawable *drawable,
			   int x, int y, int width, int height)
{
	DiaCanvasViewItem *root;
	DrawHelper helper;
	gint handle_r = (dia_handle_size() + 1) / 2;

	D((G_GNUC_PRETTY_FUNCTION));

	root = DIA_CANVAS_VIEW (hl->canvas)->root_item;

	if (!root)
		return;

	helper.drawable = drawable;
	helper.x = x + handle_r;
	helper.y = y + handle_r;
	helper.width = width;
	helper.height = height;

	dia_canvas_view_item_foreach (root, real_draw, &helper);
}

typedef struct {
	GnomeCanvasBuf *buf;
	gdouble *affine;
} RenderHelper;

static gint
real_render (DiaCanvasViewItem *item, gpointer data)
{
	RenderHelper *helper = data;
	GnomeCanvasBuf *buf = helper->buf;
	gdouble affine[6];
	DiaCanvasItem *diaitem = item->item;
	gint i;
	GList *l;
	gdouble handle_r;
	gint handle_size;
	gint color;

	/* Groups should draw their children... but invisible items and
	 * unselected leafs should not be drawn. */
	if ((!DIA_IS_CANVAS_GROUP (item->item)
	     && !DIA_CANVAS_VIEW_ITEM_SELECT (item))
	    || !(DIA_CANVAS_ITEM_VISIBLE (diaitem)))
		return FALSE;

	handle_size = dia_handle_size ();
	handle_r = (handle_size) / 2;

	/* First draw all handles: */
	for (l = diaitem->handles, i = 0; (i < item->n_handle_pos * 2) && (l != NULL); l = l->next, i += 2) {
		DiaHandle *handle = l->data;

		color = find_handle_color (handle, item);

		art_affine_translate (affine,
				      item->handle_pos[i] - handle_r,
				      item->handle_pos[i + 1] - handle_r);
		art_affine_multiply (affine, affine, helper->affine);

		art_rgb_rgba_affine (buf->buf,
				     buf->rect.x0, buf->rect.y0,
				     buf->rect.x1, buf->rect.y1,
				     buf->buf_rowstride,
				     handle_image[color],
				     handle_size, handle_size,
				     handle_size * sizeof (guint32),
				     affine, ART_FILTER_NEAREST, NULL);
	}

	return TRUE;
}


/**
 * dia_handle_layer_render:
 * @item: 
 * @buf: 
 *
 * 
 **/
static void
dia_handle_layer_render (GnomeCanvasItem *hl, GnomeCanvasBuf *buf)
{
	RenderHelper helper;
	DiaCanvasViewItem *root;
	gdouble affine[6];

	D((G_GNUC_PRETTY_FUNCTION));

	root = DIA_CANVAS_VIEW (hl->canvas)->root_item;

	if (!root)
		return;
	
	gnome_canvas_buf_ensure_buf (buf);

	gnome_canvas_item_i2w_affine (hl, affine);

	helper.buf = buf;
	helper.affine = affine;

	dia_canvas_view_item_foreach (root, real_render, &helper);
}


/**
 * recursive_find_closest_handle:
 * @item: 
 * @x: 
 * @y: 
 * @bb: 
 * @actual_handle: 
 *
 * Find the closest handle near the cursor. The cursor is placed at (x, y).
 * The distance is calculated using 'dy + dy' instead of 'dx^2 * dy^2', for
 * speed of course.
 *
 * Return value: 
 **/
static double
recursive_find_closest_handle (DiaCanvasViewItem *vitem,
	       		       gdouble x, gdouble y,
			       DiaRectangle *bb,
	       		       DiaHandle **actual_handle,
			       DiaCanvasViewItem **actual_item)
{
	gdouble wx = 0.0, wy = 0.0, dist, closest;
	GList *hlist;
	DiaHandle *handle;
	DiaCanvasViewItem *closest_item = NULL;
	DiaCanvasItem *item = vitem->item;

	closest = G_MAXDOUBLE;
	*actual_handle = NULL;
	*actual_item = NULL;

	if (DIA_IS_CANVAS_GROUP (item)) {
		GList *childs = GNOME_CANVAS_GROUP (vitem)->item_list;
		DiaCanvasViewItem *child;
		
		for (; childs != NULL; childs = childs->next) {
			child = childs->data;
			/* check if the child is a candidate for being 
			 * grabbed on a handle: */
			/*if (!(GTK_OBJECT_FLAGS (child->item) & DIA_VISIBLE)
			    || (((GnomeCanvasItem*)child)->x1 > bb->right)
			    || (((GnomeCanvasItem*)child)->x2 < bb->left)
			    || (((GnomeCanvasItem*)child)->y1 > bb->bottom)
			    || (((GnomeCanvasItem*)child)->y2 < bb->top))
				continue;
			 */
			dist = recursive_find_closest_handle (child, x, y,
							      bb, &handle,
							      &closest_item);
			if (handle && (dist <= closest)) {
				closest = dist;
				*actual_handle = handle;
				*actual_item = closest_item;
			}
		}
	}

	for (hlist = item->handles; hlist != NULL; hlist = hlist->next) {
		handle = hlist->data;

		if (!(handle->movable))
			continue;

		dia_handle_get_pos_w (handle, &wx, &wy);

		dist = fabs (wx - x) + fabs (wy - y);
		if (dist <= closest) {
			closest = dist;
			*actual_handle = handle;
			*actual_item = vitem;
		}
	}
	return closest;
}


/**
 * dia_handle_layer_point:
 * @item: 
 * @x: 
 * @y: 
 * @cx: 
 * @cy: 
 * @actual_item: 
 *
 * Determine the distance from a point to the object.
 * In this case: determine the distance from a point to a handle. If the mouse
 * pointer is close to a handle, 0.0 is returned so that the handle layer
 * will get the focus and is able to move the actual handle.
 * The closest handle is stored in DiaHandleLayer::closest_handle.
 *
 * Return value: Distance from a point to the closest handle.
 **/
static double
dia_handle_layer_point (GnomeCanvasItem *item, double x, double y,
			int cx, int cy, GnomeCanvasItem **actual_item)
{
	DiaRectangle bb;
	gdouble dist;
	gdouble r;
	DiaCanvasViewItem *root;
	DiaHandleLayer *layer;
	DiaHandle *closest_handle = NULL;
	DiaCanvasViewItem *closest_item = NULL;

	/* This action is useless if we have a grabbed object... */
	if (item->canvas->grabbed_item == item) {
		*actual_item = item;
		return 0.0;
	} else if (item->canvas->grabbed_item) {
		return G_MAXDOUBLE;
	}

	root = DIA_CANVAS_VIEW (item->canvas)->root_item;

	if (!root)
		return G_MAXDOUBLE;

	r = dia_handle_size() / (2 * dia_canvas_view_get_zoom (DIA_CANVAS_VIEW (item->canvas)));
	bb.right = bb.bottom = r;
	bb.left = bb.top = -bb.right;

	layer = DIA_HANDLE_LAYER (item);

	/* First try the focused item: */
	if (DIA_CANVAS_VIEW (item->canvas)->focus_item) {
		/* Dist is the distance from (x, y) to closest_handle. */
		dist = recursive_find_closest_handle (DIA_CANVAS_VIEW (item->canvas)->focus_item,
						      x, y, &bb,
						      &closest_handle, &closest_item);
		if (closest_handle) {
			gdouble wx, wy;

			dia_handle_get_pos_w (closest_handle, &wx, &wy);

			if ((x >= wx + bb.left) && (x <= wx + bb.right)
			    && (y >= wy + bb.top) && (y <= wy + bb.bottom)) {
				layer->closest_handle = closest_handle;
				layer->closest_item = closest_item;
				*actual_item = item;
				D(("handle on closest item"));
				return 0.0;
			}
		}
	}

	/* Try the other objects: */
	dist = recursive_find_closest_handle (root, x, y, &bb,
					      &closest_handle, &closest_item);

	if (closest_handle) {
		gdouble wx, wy;

		g_assert (DIA_IS_HANDLE (closest_handle));
		dia_handle_get_pos_w (closest_handle, &wx, &wy);

		if ((x >= wx + bb.left) && (x <= wx + bb.right)
		    && (y >= wy + bb.top) && (y <= wy + bb.bottom)) {
			layer->closest_handle = closest_handle;
			layer->closest_item = closest_item;
			*actual_item = item;
			D(("found handle!"));
			return 0.0;
		}
	}

	return G_MAXDOUBLE;
}

/**
 * dia_handle_layer_button_press:
 * @item: 
 * @event: 
 * 
 * 
 * 
 * Return value: 
 **/
static gboolean
dia_handle_layer_button_press (GnomeCanvasItem *item, GdkEventButton *event)
{
	DiaHandleLayer *handle_layer = (DiaHandleLayer*) item;

	if ((event->button == 1)
	    && (handle_layer->closest_handle)) {
		if (!(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)))
			dia_canvas_view_unselect_all (DIA_CANVAS_VIEW (item->canvas));

		handle_layer->grabbed_handle = handle_layer->closest_handle;
		handle_layer->event_mask = (DiaEventMask) event->state;

		dia_handle_remove_all_constraints (handle_layer->grabbed_handle);

		handle_layer->cp_item = handle_layer->grabbed_handle->connected_to;

		dia_canvas_item_request_update (DIA_CANVAS_VIEW_ITEM (handle_layer->closest_item)->item);

		gnome_canvas_item_grab (item, DIA_HANDLE_EVENTS,
					NULL, GDK_CURRENT_TIME);
					//NULL, event->time);

		dia_canvas_view_focus (DIA_CANVAS_VIEW (item->canvas),
				       handle_layer->closest_item);
		return TRUE;
	}
	return FALSE;
}

static gboolean
dia_handle_layer_button_release (GnomeCanvasItem *item, GdkEventButton *event)
{
	DiaHandleLayer *handle_layer = (DiaHandleLayer*) item;

	if (handle_layer->grabbed_handle
	    && (event->button == 1)) {
		/* Disconnect only if the object to connect to is
		 * not the object it is already connected to. */
		if (handle_layer->grabbed_handle->connected_to
		    && handle_layer->grabbed_handle->connected_to != handle_layer->cp_item) {
			dia_canvas_item_disconnect (handle_layer->grabbed_handle->connected_to,
						    handle_layer->grabbed_handle);
		}
		if (handle_layer->cp_item) {
			dia_canvas_item_connect (handle_layer->cp_item,
						 handle_layer->grabbed_handle);
			handle_layer->cp_item = NULL;
		}
		
		/* Update the handle's owner always: */
		dia_canvas_item_request_update (handle_layer->grabbed_handle->owner);

		dia_handle_layer_request_redraw_handle (handle_layer,
					handle_layer->grabbed_handle);

		//gnome_canvas_item_ungrab (item, event->time);
		gnome_canvas_item_ungrab (item, GDK_CURRENT_TIME);
		handle_layer->grabbed_handle = NULL;
		return TRUE;
	}
	return FALSE;
}

/* If a handle is moved several things should happen in a
 * specific order:
 * 1. Ask the owner object if the movement is allowed at all
 * 2. Find a connection point nearby the handle
 * 3. If no CP is found, snap to the grid (if enabled)
 * 4. Ask the owner again, you can't be too sure.
 * 5. Move the handle.
 */
static gboolean
dia_handle_layer_motion_notify (GnomeCanvasItem *item, GdkEventMotion *event)
{
	DiaHandleLayer *handle_layer = (DiaHandleLayer*) item;

	if (handle_layer->grabbed_handle
	    && (event->state & GDK_BUTTON1_MASK)) {
		gdouble glue_x, glue_y, dist = G_MAXDOUBLE;
		gdouble motion_x = event->x;
		gdouble motion_y = event->y;

		/* 2. Find a connection point nearby the handle */
		if (handle_layer->grabbed_handle->connectable) {
			dist = dia_canvas_glue_handle
				(handle_layer->grabbed_handle->owner->canvas,
				 handle_layer->grabbed_handle,
				 motion_x, motion_y,
				 &glue_x, &glue_y,
				 (DiaCanvasItem**) &handle_layer->cp_item);
		}
		/* 3. If no CP is found, snap to the grid (if enabled). */
		if (!handle_layer->cp_item
		    || (dist > handle_layer->glue_distance)) {
			handle_layer->cp_item = NULL;
			glue_x = motion_x;
			glue_y = motion_y;

			dia_canvas_snap_to_grid (DIA_CANVAS_VIEW (item->canvas)->canvas, &glue_x, &glue_y);
		}
		
		/* 4. Ask the handle's owner if the move is allowed */
		if (DIA_CANVAS_ITEM_GET_CLASS (handle_layer->grabbed_handle->owner)->handle_motion)
			DIA_CANVAS_ITEM_GET_CLASS (handle_layer->grabbed_handle->owner)->handle_motion
				(handle_layer->grabbed_handle->owner,
				 handle_layer->grabbed_handle,
				 &glue_x, &glue_y,
				 handle_layer->event_mask);
		
		/* 5.Moving handle and request regions to be redrawn: */
		dia_handle_layer_request_redraw_handle (handle_layer,
					handle_layer->grabbed_handle);
		
		dia_handle_set_pos_w (handle_layer->grabbed_handle,
				      glue_x, glue_y);

		dia_handle_layer_request_redraw_handle (handle_layer,
					handle_layer->grabbed_handle);
		return TRUE;
	}
	return FALSE;
}

static gboolean
dia_handle_layer_event (GnomeCanvasItem *item, GdkEvent *event)
{
	switch (event->type) {
	case GDK_BUTTON_PRESS:
		return dia_handle_layer_button_press (item, &event->button);
	case GDK_BUTTON_RELEASE:
		return dia_handle_layer_button_release (item, &event->button);
	case GDK_MOTION_NOTIFY:
		return dia_handle_layer_motion_notify (item, &event->motion);
	default:
		break;
	}
	return FALSE;
}

/**
 * dia_handle_layer_update_handles:
 * @layer: 
 * @item: 
 *
 * Redraw handles if nessesary. The old handle positions are stored
 * in the #DiaCanvasViewItem to ensure proper redrawing.
 **/
void
dia_handle_layer_update_handles (DiaHandleLayer *layer, DiaCanvasViewItem *item)
{
	DiaCanvasItem *diaitem = item->item;
	int i;
	GList *l;

	/* Create a completely new list of handle positions on the
	 * DiaCanvasViewItem if the lengths of the lists differ. */
	if (item->n_handle_pos != g_list_length (diaitem->handles)) {

		/* Request redraw for the old handle positions: */
		for (i = 0; i < item->n_handle_pos * 2; i += 2) {
			dia_handle_layer_request_redraw (layer,
							 item->handle_pos[i],
							 item->handle_pos[i+1]);
		}

		/* Now reconstruct the new (old) handle_coordinates: */
		item->n_handle_pos = g_list_length (diaitem->handles);
		item->handle_pos = g_realloc (item->handle_pos,
					      item->n_handle_pos
					      * sizeof (gdouble) * 2);
		
		for (i = 0, l = diaitem->handles; l != NULL; i += 2, l = l->next) {
			dia_handle_layer_get_pos_c (layer, DIA_HANDLE (l->data),
						    &item->handle_pos[i],
						    &item->handle_pos[i+1]);

			dia_handle_layer_request_redraw (layer,
							 item->handle_pos[i],
							 item->handle_pos[i+1]);
		}
	} else {
		/* Update the position matrix on the canvasviewitem */
		for (i = 0, l = diaitem->handles; l != NULL; i += 2, l = l->next) {
			dia_handle_layer_request_redraw (layer,
							 item->handle_pos[i],
							 item->handle_pos[i+1]);
			dia_handle_layer_get_pos_c (layer, DIA_HANDLE (l->data),
						    &item->handle_pos[i],
						    &item->handle_pos[i+1]);
			dia_handle_layer_request_redraw (layer,
							 item->handle_pos[i],
							 item->handle_pos[i+1]);
		}
	}
}

/**
 * dia_handle_layer_get_pos_c:
 * @layer: 
 * @handle: 
 * @x: 
 * @y: 
 *
 * Given a handle, return the position of that handle in canvas coordinates.
 **/
void
dia_handle_layer_get_pos_c (DiaHandleLayer *layer, DiaHandle *handle,
			    gint *x, gint *y)
{
	GnomeCanvasItem *gitem = GNOME_CANVAS_ITEM (layer)->canvas->root;
	gdouble wx, wy;
	
	dia_handle_get_pos_w (handle, &wx, &wy);

	gnome_canvas_w2c (gitem->canvas, wx, wy, x, y);
}


/**
 * dia_handle_layer_request_redraw:
 * @layer: 
 * @x: Position of the handle, in canvas coordinates
 * @y: 
 *
 * Ask the canvas to do a redraw of the a handle. Only the handles @x and @y
 * position need to be given (in canvas coordinates).
 **/
void
dia_handle_layer_request_redraw (DiaHandleLayer *layer, gint x, gint y)
{
	ArtDRect r1, r2;
	ArtIRect ir;
	gdouble affine[6];
	gdouble r = (((gdouble) dia_handle_size ()) / 2.0)  + 0.5;

	g_return_if_fail (DIA_IS_HANDLE_LAYER (layer));

	gnome_canvas_item_i2w_affine ((GnomeCanvasItem*) layer, affine);

	r1.x0 = x - r;
	r1.y0 = y - r;
	r1.x1 = x + r;
	r1.y1 = y + r;
	/* x1 = (gint) (x - r);
	y1 = (gint) (y - r);
	x2 = (gint) (x + r);
	y2 = (gint) (y + r); */
	art_drect_affine_transform (&r2, &r1, affine);
	art_drect_to_irect (&ir, &r2);
	gnome_canvas_request_redraw (GNOME_CANVAS_ITEM (layer)->canvas,
				     ir.x0, ir.y0, ir.x1, ir.y1);
				     //x1, y1, x2, y2);
}

/**
 * dia_handle_layer_request_redraw_handle:
 * @layer: 
 * @handle: 
 *
 * Do a redraw for @handle.
 **/
void
dia_handle_layer_request_redraw_handle (DiaHandleLayer *layer,
					DiaHandle *handle)
{
	gint x, y;

	/* TODO: try doing a lookup in the handle's handle_pos list if possible
	 */
	dia_handle_layer_get_pos_c (layer, handle, &x, &y);
	dia_handle_layer_request_redraw (layer, x, y);
}

/**
 * dia_handle_layer_grab_handle:
 * @layer: 
 * @handle: 
 *
 * Grab a handle. The handle should be grabbed suring a button-press event,
 * since the handle-layer will grab the widgets events.
 **/
void
dia_handle_layer_grab_handle (DiaHandleLayer *layer,
			      DiaHandle *handle)
{
	DiaCanvasViewItem *vitem = NULL;

	g_return_if_fail (DIA_IS_HANDLE_LAYER (layer));
	g_return_if_fail (DIA_IS_HANDLE (handle));

	layer->grabbed_handle = handle;
	layer->event_mask = 0;

	dia_handle_remove_all_constraints (layer->grabbed_handle);

	layer->cp_item = layer->grabbed_handle->connected_to;

	dia_canvas_item_request_update (layer->grabbed_handle->owner);

	gnome_canvas_item_grab (GNOME_CANVAS_ITEM (layer),
				DIA_HANDLE_EVENTS, NULL, GDK_CURRENT_TIME);

	vitem = dia_canvas_view_find_view_item (DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (layer)->canvas),
						layer->grabbed_handle->owner);
	g_return_if_fail (DIA_IS_CANVAS_VIEW_ITEM (vitem));
	
	dia_canvas_view_focus (DIA_CANVAS_VIEW (GNOME_CANVAS_ITEM (layer)->canvas),
			       vitem);
}

